from typing import Literal
from pyspedfiscal.campos import CampoAlphanumerico, CampoEnumerate
from pyspedfiscal.models import Registro


class Oper(CampoEnumerate):
    """Indicador do tipo de operação"""
    combustiveis_lubrificantes = '0'
    leasing_veiculos_faturamento_direto = '1'


class RegistroC105(Registro):
    """
    OPERAÇÕES COM ICMS ST RECOLHIDO PARA UF DIVERSA DO
    DESTINATÁRIO DO DOCUMENTO FISCAL (CÓDIGO 55)
    """
    reg: Literal['C105']
    oper: Oper
    uf: CampoAlphanumerico



from typing import List
from ..models import Bloco
from .registro_0000 import Registro0000
from .registro_0001 import Registro0001
from .registro_0200 import Registro0200
from .registro_0460 import Registro0460


class Bloco0(Bloco):
    def __init__(self):
        self.registro_0000: Registro0000 = None
        self.registro_0001: Registro0001 = None

    @classmethod
    def ler_registros(cls, registros: List[str]) -> 'Bloco0':
        bloco = cls()

        for registro in registros:
            tipo_registro = registro[1:5].upper()

            # Nivel 0 - Abertura do Arquivo Digital
            if tipo_registro == '0000':
                bloco.registro_0000 = Registro0000.ler_registro(registro)

            # Nivel 1 - Registro Pai
            if tipo_registro == '0001':
                bloco.registro_0001 = Registro0001.ler_registro(registro)

            # Nivel 2 - Filho de 0001
            if tipo_registro == '0200':
                bloco.registro_0001.registros_0200.append(
                    Registro0200.ler_registro(registro)
                )

            # Nivel 2 - Filho de 0001
            if tipo_registro == '0460':
                bloco.registro_0001.registros_0460.append(
                    Registro0460.ler_registro(registro)
                )

        return bloco



from typing import Literal
from pyspedfiscal.models import Registro
from pyspedfiscal.campos import CampoEnumerate, CampoData, CampoAlphanumerico, CampoInteiro, CampoCNPJ, CampoCPF
from pyspedfiscal.tabelas import CodVer


class CodFin(CampoEnumerate):
    original = '0'
    substituto = '1'


class IndAtiv(CampoEnumerate):
    industrial = '0'
    outros = '1'


class Registro0000(Registro):
    """ Abertura do Arquivo Digital """

    reg: Literal['0000']
    cod_ver: CodVer
    cod_fin: CodFin
    dt_ini: CampoData
    dt_fin: CampoData
    nome: CampoAlphanumerico
    cnpj: CampoCNPJ
    cpf: CampoCPF
    uf: CampoAlphanumerico
    ie: CampoAlphanumerico
    cod_mun: CampoInteiro
    im: CampoAlphanumerico
    suframa: CampoAlphanumerico
    ind_perfil: CampoAlphanumerico
    ind_ativ: IndAtiv



from typing import Literal
from pyspedfiscal.campos import CampoAlphanumerico, CampoEnumerate, CampoSerie, CampoInteiro, CampoData, CampoChaveAcesso
from pyspedfiscal.tabelas import CodMod
from pydantic import Field
from pyspedfiscal.models import Registro


class IndOper(CampoEnumerate):
    """ Indicador do tipo de operação """
    entrada = '0'
    saida = '1'


class IndEmit(CampoEnumerate):
    """ Indicador do emitente do documento fiscal """
    emissao_propria = '0'
    terceiros = '1'


class RegistroC113(Registro):
    """DOCUMENTO FISCAL REFERENCIADO"""
    reg: Literal['C113']
    ind_oper: IndOper
    ind_emit: IndEmit
    cod_part: CampoAlphanumerico
    cod_mod: CodMod
    ser: CampoSerie
    sub: CampoInteiro
    num_doc: CampoInteiro
    dt_doc: CampoData
    chv_doce: CampoChaveAcesso



from typing import List, Literal
from pyspedfiscal.models import Registro
from pyspedfiscal.campos import CampoAlphanumerico, CampoEnumerate
from pydantic import Field
from .registro_e100 import RegistroE100


class IndMov(CampoEnumerate):
    """ Indicador de movimento """
    com_movimento = '0'
    sem_movimento = '1'


class RegistroE001(Registro):
    """ ABERTURA DO BLOCO E """
    reg: Literal['E001']
    ind_mov: IndMov

    # Registros filhos nivel 2
    registros_e100: List[RegistroE100] = Field(
        default_factory=list, exclude=True)



from typing import Literal
from pyspedfiscal.campos import CampoAlphanumerico, CampoEnumerate
from pydantic import Field
from pyspedfiscal.models import Registro


class IndProc(CampoEnumerate):
    """Indicador da origem do processo"""
    sefaz = '0'
    justica_federal = '1'
    justica_estadual = '2'
    secex_srf = '3'
    outros = '9'


class RegistroC111(Registro):
    """PROCESSO REFERENCIADO"""
    reg: Literal['C111']
    num_proc: CampoAlphanumerico
    ind_proc: IndProc



from typing import List, Literal
from pyspedfiscal.models import Registro
from pyspedfiscal.campos import CampoData
from pydantic import Field
from .registro_e110 import RegistroE110


class RegistroE100(Registro):
    """ PERÍODO DA APURAÇÃO DO ICMS """
    reg: Literal['E100']
    dt_ini: CampoData
    dt_fin: CampoData

    # Registros filhos nivel 3
    registros_e110: List[RegistroE110] = Field(
        default_factory=list, exclude=True)



from typing import List
from ..models import Bloco
from .registro_9001 import Registro9001
from .registro_9999 import Registro9999


class Bloco9(Bloco):
    def __init__(self):
        self.registro_9001: Registro9001 = None
        self.registro_9999: Registro9999 = None

    @classmethod
    def ler_registros(cls, registros: List[str]) -> 'Bloco9':
        bloco = cls()

        for registro in registros:
            tipo_registro = registro[1:5].upper()

            # Nivel 1 - Registro Pai
            if tipo_registro == '9001':
                bloco.registro_9001 = Registro9001.ler_registro(registro)

            # Nivel 0 - Encerramento do arquivo
            # Ultimo Registro do arquivo
            elif tipo_registro == '9999':
                bloco.registro_9999 = Registro9999.ler_registro(registro)

        return bloco



from typing import Literal
from pyspedfiscal.campos import CampoDecimal
from pyspedfiscal.models import Registro


class RegistroC101(Registro):
    """
    INFORMAÇÃO COMPLEMENTAR DOS DOCUMENTOS FISCAIS
    QUANDO DAS OPERAÇÕES INTERESTADUAIS DESTINADAS A CONSUMIDOR FINAL
    NÃO CONTRIBUINTE EC 87/15 (CÓDIGO 55)
    """
    reg: Literal['C101']
    vl_fcp_uf_dest: CampoDecimal
    vl_icms_uf_dest: CampoDecimal
    vl_icms_uf_rem: CampoDecimal



from pyspedfiscal.campos import CampoEnumerate


class CodSit(CampoEnumerate):
    """ Tabela 4.1.2 - Código da Situação do Documento """
    regular = '00'
    extemporaneo = '01'
    cancelado = '02'
    cancelado_extemporaneo = '03'
    denegado = '04'
    inutilizado = '05'
    complementar = '06'
    complementar_extemporaneo = '07'
    regime_especial = '08'


class CodMod(CampoEnumerate):
    """ Tabela 4.1.1 - Tabela Documentos Fiscais do ICMS """
    nota_fiscal = '01'
    nota_fiscal_avulsa = '1B'
    nota_fiscal_de_venda_consumidor = '02'
    cupom_fiscal = '2D'
    cupom_fiscal_passagem = '2E'
    nota_fiscal_produtor = '04'
    energia_eletrica = '06'
    nota_fiscal_transporte = '07'
    conhecimento_de_transporte = '08'
    conhecimento_de_transporte_avulso = '8B'
    conhecimento_de_transporte_aquaviario = '09'
    conhecimento_aereo = '10'
    conhecimento_de_transporte_ferroviario = '11'
    bilhete_de_passagem_rodoviario = '13'
    bilhete_de_passagem_aquaviario = '14'
    bilhete_de_passagem_nota_bagagem = '15'
    bilhete_de_passagem_ferroviario = '16'
    resumo_de_movimento_diario = '18'
    nota_fiscal_servico_de_comunicacao = '21'
    nota_fiscal_servico_de_telecomunicacao = '22'
    conhecimento_de_transporte_multimodal = '26'
    nota_fiscal_transporte_ferroviario = '27'
    nota_fiscal_gas_canalizado = '28'
    nota_Fiscal_agua_canalizada = '29'
    nota_fiscal_eletronica = '55'
    conhecimento_de_transporte_eletronico = '57'
    cupom_fiscal_eletronico_sat = '59'
    cupom_fiscal_eletronico_secf = '60'
    bilhete_de_passagem_eletronico = '63'
    nota_fiscal_eletronica_consumidor_final = '65'
    conhecimento_de_transporte_eletronico_os = '67'


class CodVer(CampoEnumerate):
    """ Tabela 3.1.1 - Tabela Versão do Leiaute - Sped Fiscal """
    layout_1_01 = '002'
    layout_1_02 = '003'
    layout_1_03 = '004'
    layout_1_04 = '005'
    layout_1_05 = '006'
    layout_1_06 = '007'
    layout_1_07 = '008'
    layout_1_08 = '009'
    layout_1_09 = '010'
    layout_1_10 = '011'
    layout_1_11 = '012'
    layout_1_12 = '013'
    layout_1_13 = '014'
    layout_1_14 = '015'
    layout_1_15 = '016'


class GenItemMercServ(CampoEnumerate):
    """ Tabela 4.2.1 - Tabela Gênero do Item de Mercadoria/Serviço """
    pass



from .bloco_0 import Bloco0
from .bloco_b import BlocoB
from .bloco_c import BlocoC
from .bloco_d import BlocoD
from .bloco_e import BlocoE
from .bloco_g import BlocoG
from .bloco_h import BlocoH
from .bloco_k import BlocoK
from .bloco_1 import Bloco1
from .bloco_9 import Bloco9

import re
from pyspedfiscal.exception import ArquivoInvalido


class SpedFiscal:
    def __init__(self) -> None:
        self.bloco_0: Bloco0 = None
        self.bloco_b: BlocoB = None
        self.bloco_c: BlocoC = None
        self.bloco_d: BlocoD = None
        self.bloco_e: BlocoE = None
        self.bloco_g: BlocoG = None
        self.bloco_h: BlocoH = None
        self.bloco_k: BlocoK = None
        self.bloco_1: Bloco1 = None
        self.bloco_9: Bloco9 = None

    def validar(self, arquivo: str):

        bloco_0000_pattern = r'^(\|0000\|)(\d){3,}\|[0,1]\|((\d){8,}\|){2}(.{0,100})\|(\d){,14}\|(\d){,11}\|.{2,}\|.{0,14}\|(\d){0,7}\|.*\|.{0,9}\|[a,b,c,A,B,C]\|[0,1]\|$'

        try:
            with open(arquivo, 'r', encoding='latin-1') as arq_sped:
                bloco_0000 = next(arq_sped)
        except OSError as e:
            raise OSError("Não foi possível ler o arquivo:\n", repr(e))
        except Exception as e:
            raise Exception(
                "Erro inesperado, entre em contato com o suporte:\n", repr(e))
        else:
            return bool(re.match(bloco_0000_pattern, bloco_0000))

    @classmethod
    def importar(cls, arquivo: str) -> 'SpedFiscal':
        """Importa um arquivo Sped

        Args:
            arquivo (str): Arquivo Sped

        Raises:
            ArquivoInvalido: Erros de validação no arquivo
        """

        sped = cls()

        # Valida o arquivo
        if not sped.validar(arquivo):
            raise ArquivoInvalido(arquivo)

        with open(arquivo, 'r', encoding='latin-1') as arq_sped:

            resgitros = arq_sped.readlines()

            # Remove as linhas que não são registro
            resgitros = [
                registro for registro in resgitros if registro.startswith('|')
            ]

            sped.bloco_0 = Bloco0.ler_registros(resgitros)
            sped.bloco_b = BlocoB.ler_registros(resgitros)
            sped.bloco_c = BlocoC.ler_registros(resgitros)
            sped.bloco_d = BlocoD.ler_registros(resgitros)
            sped.bloco_e = BlocoE.ler_registros(resgitros)
            sped.bloco_g = BlocoG.ler_registros(resgitros)
            sped.bloco_h = BlocoH.ler_registros(resgitros)
            sped.bloco_k = BlocoK.ler_registros(resgitros)
            sped.bloco_1 = Bloco1.ler_registros(resgitros)
            sped.bloco_9 = Bloco9.ler_registros(resgitros)

        return sped



from pathlib import Path


class SpedError(Exception):
    pass


class ArquivoInvalido(SpedError):
    def __init__(self, arquivo: str) -> None:
        nome_arquivo = Path(arquivo).name
        self._message = f'o arquivo {nome_arquivo} não é um Sped FIscal válido.'

    def __repr__(self) -> str:
        return self._message

    def __str__(self) -> str:
        return self._message


class TamanhoRegistroInvalido(SpedError):
    pass


class LinhaRegistroInvalida(SpedError):
    pass


class ValorCampoInvalido(SpedError):
    def __init__(self, campo: str, valor: str) -> None:
        self._campo = campo
        self._valor = valor
        self._message = f'O valor {self._valor} não é um valor valido para o campo {self._campo}.'

    def __repr__(self) -> str:
        return self._message

    def __str__(self) -> str:
        return self._message



from typing import List, Literal
from pyspedfiscal.campos import CampoAlphanumerico
from pydantic import Field
from pyspedfiscal.models import Registro
from pyspedfiscal.bloco_c.registro_c111 import RegistroC111
from pyspedfiscal.bloco_c.registro_c112 import RegistroC112
from pyspedfiscal.bloco_c.registro_c113 import RegistroC113
from pyspedfiscal.bloco_c.registro_c114 import RegistroC114
from pyspedfiscal.bloco_c.registro_c115 import RegistroC115
from pyspedfiscal.bloco_c.registro_c116 import RegistroC116


class RegistroC110(Registro):
    """
    INFORMAÇÃO COMPLEMENTAR DA NOTA FISCAL (CÓDIGO 01, 1B, 04 e 55)
    """
    reg: Literal['C110']
    cod_inf: CampoAlphanumerico
    txt_compl: CampoAlphanumerico

    registros_c111: List[RegistroC111] = Field(
        default_factory=list, exclude=True)
    registros_c112: List[RegistroC112] = Field(
        default_factory=list, exclude=True)
    registros_c113: List[RegistroC113] = Field(
        default_factory=list, exclude=True)
    registros_c114: List[RegistroC114] = Field(
        default_factory=list, exclude=True)
    registros_c115: List[RegistroC115] = Field(
        default_factory=list, exclude=True)
    registros_c116: List[RegistroC116] = Field(
        default_factory=list, exclude=True)



from typing import List, Literal
from pyspedfiscal.campos import CampoDecimal, CampoInteiro, CampoAlphanumerico
from pydantic import Field
from pyspedfiscal.models import Registro


class RegistroC190(Registro):
    """ Registro Analítico do Documento (código 01, 1B, 04, 55 e 65)  """
    reg: Literal['C190']
    cst_icms: CampoInteiro
    cfop: CampoInteiro
    aliq_icms: CampoDecimal
    vl_opr: CampoDecimal
    vl_bc_icms: CampoDecimal
    vl_icms: CampoDecimal
    vl_bc_icms_st: CampoDecimal
    vl_icms_st: CampoDecimal
    vl_red_bc: CampoDecimal
    vl_ipi: CampoDecimal
    cod_obs: CampoAlphanumerico



from typing import Literal
from pyspedfiscal.campos import CampoDecimal, CampoAlphanumerico
from pyspedfiscal.models import Registro


class RegistroC197(Registro):
    """ 
    Outras Obrigações Tributárias, Ajustes e Informações provenientes de
    Documento Fiscal 
    """
    reg: Literal['C197']
    cod_aj: CampoAlphanumerico
    descr_compl_aj: CampoAlphanumerico
    cod_item: CampoAlphanumerico
    vl_bc_icms: CampoDecimal
    aliq_icms: CampoDecimal
    vl_icms: CampoDecimal
    vl_outros: CampoDecimal



from typing import Literal
from pyspedfiscal.campos import CampoData, CampoInteiro, CampoChaveAcesso
from pyspedfiscal.models import Registro
from pyspedfiscal.tabelas import CodMod


class RegistroC116(Registro):
    """CUPOM FISCAL ELETRÔNICO REFERENCIADO"""
    reg: Literal['C116']
    cod_mod: CodMod
    nr_sat: CampoInteiro
    chv_cfe: CampoChaveAcesso
    num_cfe: CampoInteiro
    dt_doc: CampoData



from typing import List
from ..models import Bloco


class BlocoG(Bloco):
    def __init__(self):
        pass

    @classmethod
    def ler_registros(cls, registros: List[str]) -> 'BlocoG':
        bloco = cls()

        for registro in registros:
            tipo_registro = registro[1:5].upper()

        return bloco



from typing import Literal
from pyspedfiscal.models import Registro
from pyspedfiscal.campos import CampoEnumerate
from typing import List
from pydantic import Field

from .registro_0200 import Registro0200
from .registro_0460 import Registro0460


class IndMov(CampoEnumerate):
    """ Indicador de movimento """
    com_movimento = '0'
    sem_movimento = '1'


class Registro0001(Registro):
    """ ABERTURA DO BLOCO 0 """

    reg: Literal['0001']
    ind_mov: IndMov

    # Registros filhos nivel 2
    registros_0200: List[Registro0200] = Field(
        default_factory=list, exclude=True)
    registros_0460: List[Registro0460] = Field(
        default_factory=list, exclude=True)



from datetime import date
from decimal import Decimal
from pyspedfiscal.bloco_0 import Bloco0
import unittest


class TestBloco0(unittest.TestCase):

    def setUp(self) -> None:

        bloco = """
|0000|016|0|01072022|31072022|EMPRESA TESTE|14028298000152||SP|009371438460|3550308|||A|0|
|0001|0|
|0002|00|
|0005|EMPRESA TESTE|25821270|RUA EMPRESA TESTE|1111||BAIRRO EMPRESA TESTE|2499999999|||
|0100|CONTADOR TESTE|25142702000|1-SP-111111/O-1||25805290|RUA CONTADOR|111|ENDERECO CONTADOR|BAIRRO CONTADOR|2499999999||emailcontador@contador.com.br|3550308|
|0150|1|FORNECEDOR 1|1058|03446003000141|||3301702||ENDERECO FORNECEDOR 1|1111||BAIRRO FORNECEDOR 1|
|0150|2|FORNECEDOR 2|1058|97509345000107|||3550308||ENDERECO FORNECEDOR 1|2222||BAIRRO FORNECEDOR 2|
|0150|3|FORNECEDOR 3|1058|05929208000103|||3550308||ENDERECO FORNECEDOR 1|3333||BAIRRO FORNECEDOR 3|
|0150|4|FORNECEDOR 4|1058|13992567000133|||3550308||ENDERECO FORNECEDOR 1|4444||BAIRRO FORNECEDOR 4|
|0150|5|FORNECEDOR 5|1058|38284654000167|||3550308||ENDERECO FORNECEDOR 1|5555||BAIRRO FORNECEDOR 5|
|0190|KG|KILO|
|0190|UN|UNIDADE|
|0200|1|PRODUTO 1|||UN|00|27101921||||||
|0200|2|PRODUTO 2|||KG|00|27101921||||||
|0200|3|PRODUTO 3|||KG|00|27101921||||||
|0200|4|PRODUTO 4|||KG|00|27101921||||||
|0200|5|PRODUTO 5|||KG|00|27101921||||||
|0400|1653002|Natureza Teste|
|0450|1|Informacaoo Complementar Teste|
|0460|1|Observacao Teste|
|0990|21|""".splitlines()

        self.bloco_0 = Bloco0.ler_registros(bloco)

    def test_deve_ler_um_registro_0000(self):
        registro_0000 = self.bloco_0.registro_0000

        reg = registro_0000.reg
        dt_ini = registro_0000.dt_ini
        dt_fin = registro_0000.dt_fin

        self.assertEqual(reg, '0000')
        self.assertEqual(dt_ini, date(2022, 7, 1))
        self.assertEqual(dt_fin, date(2022, 7, 31))

    def test_deve_ler_um_registro_0001(self):
        registro_0001 = self.bloco_0.registro_0001

        reg = registro_0001.reg
        ind_mov = registro_0001.ind_mov

        # |0001|0|

        self.assertEqual(reg, '0001')
        self.assertEqual(ind_mov, '0')

    def test_deve_ler_um_registro_0200(self):
        registro_0200 = self.bloco_0.registro_0001.registros_0200[1]

        # |0200|2|PRODUTO 2|||KG|00|27101921||||||

        self.assertEqual(registro_0200.reg, '0200')
        self.assertEqual(registro_0200.cod_item, '2')
        self.assertEqual(registro_0200.descr_item, 'PRODUTO 2')
        self.assertEqual(registro_0200.cod_barra, '')
        self.assertEqual(registro_0200.cod_ant_item, '')
        self.assertEqual(registro_0200.unid_inv, 'KG')
        self.assertEqual(registro_0200.tipo_item, '00')
        self.assertEqual(registro_0200.cod_ncm, '27101921')
        self.assertEqual(registro_0200.ex_ipi, '')
        self.assertEqual(registro_0200.cod_gen, '')
        self.assertEqual(registro_0200.cod_lst, '')
        self.assertEqual(registro_0200.aliq_icms, '')
        self.assertEqual(registro_0200.cest, '')

    def test_deve_ler_um_registro_0460(self):
        registro_0460 = self.bloco_0.registro_0001.registros_0460[0]

        # |0460|1|Observacao Teste|

        self.assertEqual(registro_0460.reg, '0460')
        self.assertEqual(registro_0460.cod_obs, '1')
        self.assertEqual(registro_0460.txt, 'Observacao Teste')


if __name__ == '__main__':
    unittest.main()



from typing import List, Literal
from pyspedfiscal.models import Registro
from pyspedfiscal.campos import CampoDecimal
from pydantic import Field
from .registro_e115 import RegistroE115


class RegistroE110(Registro):
    """ APURAÇÃO DO ICMS – OPERAÇÕES PRÓPRIAS """

    reg: Literal['E110']
    vl_tot_debitos: CampoDecimal
    vl_aj_debitos: CampoDecimal
    vl_tot_aj_debitos: CampoDecimal
    vl_estornos_cred: CampoDecimal
    vl_tot_creditos: CampoDecimal
    vl_aj_creditos: CampoDecimal
    vl_tot_aj_creditos: CampoDecimal
    vl_estornos_deb: CampoDecimal
    vl_sld_credor_ant: CampoDecimal
    vl_sld_apurado: CampoDecimal
    vl_tot_ded: CampoDecimal
    vl_icms_recolher: CampoDecimal
    vl_sld_credor_transportar: CampoDecimal
    deb_esp: CampoDecimal

    # Registros filhos nivel 4
    registros_c115: List[RegistroE115] = Field(
        default_factory=list, exclude=True)
